/************************************************************
**
**       COPYRIGHT (C) 2005 CZECH TECHNICAL UNIVERSITY
**       COPYRIGHT (C) 2004-2005 Martin Kucera
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/* file bench_parse.c */
/* ------------------------------------------------------------------------
 * BENCH parser routines. these should track the ones in 
 * the vhdl code.
 * ------------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "includes\\parser.h"
#include "includes\\bench_format.h"

/* for compatibility with BENCH VHDL system */
#include "includes\\system_defs.h"
#include "includes\\externs.h"

void CHK_BENCH_END(void * a)
{ 
	if (a == NULL) 
    { 
		fprintf (stderr, "premature end to .bench file found\n");
		exit (1); 
    }
}

/*---------------------------------------------------------------
 * read_bench parse .bench file
 *---------------------------------------------------------------
 */

void get_inputs(char **tok_ptr, list **inputs)
{
	while (TRUE)
	{
		if (stricmp(*tok_ptr, TOKEN_INPUT))
			break;
		
		*tok_ptr = find_token (TOKEN_L_BRACKET);
		CHK_BENCH_END (*tok_ptr);
		
		*tok_ptr = get_next_token();
		CHK_BENCH_END (*tok_ptr);

//		add_port(*tok_ptr, "IN");
//		newnode(*tok_ptr);
		push(strdup(*tok_ptr), inputs);

		*tok_ptr = find_token (TOKEN_R_BRACKET);
		CHK_BENCH_END (*tok_ptr);
	
		*tok_ptr = get_next_token ();
	}
}

void get_outputs(char **tok_ptr, list **outputs)
{
	while (TRUE)
	{
		if (stricmp(*tok_ptr, TOKEN_OUTPUT))
			break;
		
		*tok_ptr = find_token (TOKEN_L_BRACKET);
		CHK_BENCH_END (*tok_ptr);
		
		*tok_ptr = get_next_token();
		CHK_BENCH_END (*tok_ptr);

//		add_port(*tok_ptr, "OUT");
//		newnode(*tok_ptr);
		push(strdup(*tok_ptr), outputs);

		*tok_ptr = find_token (TOKEN_R_BRACKET);
		CHK_BENCH_END (*tok_ptr);
	
		*tok_ptr = get_next_token ();
	}
}
struct	modules_struct
{
	char	*name;
	char	*type;
	list	*inputs;
	list	*outputs;
} ;

typedef struct modules_struct modules_info;

void get_modules(char **tok_ptr, list **modules, list **inputs, list **outputs)
{
	char	*tmp, mod_name[255], in_cnt = 0, i = 0;
	static int	cnt = 0;
	list	*inp = NULL;
	modules_info	*mod = (modules_info *)calloc(1, sizeof(modules_info));

	CHK_BENCH_END (*tok_ptr);

	tmp = strdup(*tok_ptr);
	
	/* Get module type */
	*tok_ptr = find_token (TOKEN_EQUAL);
	CHK_BENCH_END (*tok_ptr);
	
	*tok_ptr = get_next_token();
	CHK_BENCH_END (*tok_ptr);

	sprintf(mod_name, "%s.%d", *tok_ptr, cnt++);
	mod->name = strdup(mod_name);
	mod->type = strupr(strdup(*tok_ptr));

//	curobj_module = newobject(mod_name, strupr(*tok_ptr));
	
	/* Get module inputs and outputs */
	if (member_p(tmp, *inputs, stringEq) == FALSE
		&& member_p(tmp, *outputs, stringEq) == FALSE)
	{
//		add_port(tmp, "OUT");
//		newnode(tmp);
		push(strdup(tmp), outputs);
	}
	push(tmp, &mod->outputs);
//	addout(tmp, 0, 1);

	*tok_ptr = find_token (TOKEN_L_BRACKET);
	CHK_BENCH_END (*tok_ptr);

	while (strcmp(*tok_ptr, TOKEN_R_BRACKET))
	{
		in_cnt++;
		*tok_ptr = get_next_token();
		CHK_BENCH_END (*tok_ptr);

		push(strdup(*tok_ptr), &mod->inputs);
//		push(strdup(*tok_ptr), &inp);
		
		if (member_p(*tok_ptr, *inputs, stringEq) == FALSE
			&& member_p(*tok_ptr, *outputs, stringEq) == FALSE)
		{
//			add_port(*tok_ptr, "IN");
//			newnode(*tok_ptr);
			push(strdup(*tok_ptr), inputs);
		}
	
		*tok_ptr = get_next_token ();
		CHK_BENCH_END (*tok_ptr);		
	}

/*	while (inp != NULL)
	{
		tmp = (char *)pop(&inp);
		addin(tmp, i++, in_cnt);
	}
*/
	push(mod, modules);
}

void read_bench (FILE *in)
{
    char	*tok_ptr;
    char	signame[128], sigdir[32];
    char	busname[128];
    int 	high, low;
	list	*inputs = NULL, *outputs = NULL, *modules = NULL, *l, *ll;
	int		done, pos, cnt;
    
	
    in_file = in;
	
    /* before we start add the predefined constant signals 
	* $0 $1, and _DUMMY_SIGNAL_
	*/
	newnode("$0");
    newnode("$1");
    newnode("_DUMMY_SIGNAL_");

	/* comment */
	tok_ptr = get_next_token();
	while (!stricmp(tok_ptr, TOKEN_HASH))
	{
		while (TRUE)
		{
			
			if (getc(in_file) == '\n')
				break;
		}
		tok_ptr = get_next_token();
    }
    
	/* we pick up the external port info */
	get_inputs(&tok_ptr, &inputs);
	get_outputs(&tok_ptr, &outputs);

	while (tok_ptr != NULL)
	{
		get_modules(&tok_ptr, &modules, &inputs, &outputs);
		tok_ptr = get_next_token();
	}
	
	for (l = inputs; l != NULL; l = l->next)
	{
		add_port((char*)l->object, "IN");
	}
	for (l = outputs; l != NULL; l = l->next)
	{
		add_port((char*)l->object, "OUT");
	}

	for (l = inputs; l != NULL; l = l->next)
	{
		newnode(l->object);
	}
	
	for (l = outputs; l != NULL; l = l->next)
	{
		newnode(l->object);
	}

/*	while (outputs != NULL)
	{

		free(outputs->object);
		pop(&outputs);
	}
	while (inputs != NULL)
	{
		free(inputs->object);
		pop(&inputs);
	}
*/
	for (l = modules; l != NULL; l = l->next)
	{
		curobj_module = newobject(((modules_info*)l->object)->name, ((modules_info*)l->object)->type);

		adddelay(30);
		pos = 0;
		cnt = list_length(((modules_info*)l->object)->inputs);
		for (ll = ((modules_info*)l->object)->inputs, pos = 0; ll != NULL; ll = ll->next, pos++)
		{
			addin((char*)ll->object, pos, cnt);
		}
	
		pos = 0;
		cnt = list_length(((modules_info*)l->object)->outputs);
		for (ll = ((modules_info*)l->object)->outputs, pos = 0; ll != NULL; ll = ll->next, pos++)
		{
			addout((char*)ll->object, pos, cnt);
		}
		
		addpositions(curobj_module
			, list_length(((modules_info*)l->object)->inputs)
			, 0
			, cnt);
	}


//	free_list(outputs);
//	free_list(inputs);
}
