
#ifndef PLACE_H
#define PLACE_H

/*---------------------------------------------------------------
* Global (to place.c)  Declarations
*---------------------------------------------------------------
*/

static double xgr = 0; /* center of "gravity" of current box */
static double ygr = 0;
static double xgbr = 0; /* center of "gravity" of currently placed boxes */
static double ygbr = 0;

static int xSideCount;   /* number of conections for the dominant count in 
                            "best_side..." */
static int ySideCount;    /* number of conections for the next highest count in 
                            "best_side..." */

static ctree *c_tree;
/* the stem upon which the cluster tree is built and maintained. */

static int clus1 = 0, clus2 = 0;
/* identifies the clusters that should be combined in the next clustering pass*/

static clist *c_list, *partition_tree;
/* lists of module numbers, and other stats needed for the clustering calculations */

static int d1 = 0, d2 = 0;
/* used to evaluate how cross-coupled modules/partitions are placed; <d1> rep-
 * resents the depth from <m1> to <m2>, and <d2> the reverse.  (see cross_coupled_p())
 */

int rpl_merge_row[MAX_MOD];	/* Used to restore the connection matrix if a */
int rpl_scratch_row[MAX_MOD];	/* partition is clipped from the tree */
int rpl_merge_col[MAX_MOD];
int rpl_scratch_col[MAX_MOD];


/*--------------------------------------------------------------- 
 * Forward declarations defined in place.c
 *---------------------------------------------------------------
 */

mlist *trace_path();
void init_string_arrays(int p);
module *select_next_box();
int accept_node1(), accept_node2(), accept_node3(), accept_node4();
void make_part_assignments(int p, ctree *cnode);
void reset_connected_matrix(int merge, int scratch, int cCount, clist **actives
							, int *col, int *row, int how);
int interpret_block_assignments();
void forward_build_connections();
void build_connections_amongst_partitions(int p, int how);
void build_clusters();
void init_clusters(int m_count, clist **lst);
void init_string(int p);
void place_box(mlist *box, int part);
void initialize_partition_markings(int pc);
void place_partition(int part, int order);
void set_gravity_part(int p, int xb, int yb, int xbb, int ybb, int count);
void fixup_string(int part, int xflr, int yflr);
void move_string(int part, int xflr, int yflr);
void reset_borders(int xflr, int yflr);
void fixup_partition(int part, int xflr, int yflr);
void rotate_module(module *m, int rot);
void build_connections(int how);
void unsigned_build_connections();
void reverse_build_connections();
void count_inout(int m, int pn);
void reset_info();
void init_info();
void install_modules();
void assign_partitions(int pnum);
void reset_connections(int cl_count, int merge_col, int scratch_col);
void merge_clusters(int c1, int c2, int cCount, clist **t, int (*acceptance_fn)());
void set_flag_value(mlist *string, int flag_val);
void reset_path_trace(mlist *path);
void overlay_connections(int cl_count, int how);
int find_next_cluster(int c_count);
int any_placed_p(mlist *string);
int count_connections(mlist *box, int part);
int place_head(int part);
int place_string(int part);
int is_root(module *m, int j);
int line_up_mod(module *m, term *prevt);
int best_side_box(int part, mlist *box);
int find_max_diagonal(int range);
int find_max_connections_to_P0(int range);
int best_side_part(int part);
int choose_side(int common_ins, int common_outs, int in_to, int out_of);
	

#endif // #ifndef PLACE_H