/************************************************************
**
**       COPYRIGHT (C) 1993 UNIVERSITY OF PITTSBURGH
**       COPYRIGHT (C) 1996 GANNON UNIVERSITY
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/


/* file n2a.c */
/* ------------------------------------------------------------------------
 * main file manipulation routines                         spl 7/89
 *
 * ------------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*---------------------------------------------------------------
 * External functions
 *---------------------------------------------------------------
 */

#include "includes\\externs.h"
#include "includes\\n2a.h"


void main(int argc, char *argv[])
{

/*------------------------------------------------------------------------------------
 * main: parse command line, open files, parse, place internal terminals, global route, 
 *       lay system terminals, global route, local route.
 *-------------------------------------------------------------------------------------
 */


    FILE *ivffile = NULL;
    int x1, y1, x2, y2;
    nlist *nl = NULL, *systermNets = NULL;
    mlist *modulesJustPlaced, *ml;
    mlist *systerms = NULL;
    /* for date and time */
    int collapseNullModules;
	char	name[255];

    collapseNullModules=FALSE;

	if (argc != 2)
	{
		fprintf(stderr,"spar.exe : program for Automatic schematic generation of netlist\n\
			using: spar.exe \'netlist_file.bench\'\n");
		exit(1);
	}
    
    ivffile = fopen(argv[1], "r+");
	if (ivffile == NULL) 
	{
		fprintf(stderr,"can't open %s\n", ivffile);
		exit(1);
    }
    else 
		fprintf(stderr,"%s, used for ivffile\n", argv[1]);
    
	sprintf(name, "%s.ps", argv[1]);
    lroutputFile = fopen(name, "w+");
   
    /*
     * main body
     */

   /* parse input */

    read_bench(ivffile);  

    /* place */
	fprintf(stderr,"Placing internal modules...");

	partition_count = partition();

	if (collapseNullModules == TRUE) 
		clip_null_gates();


	erase_systerms();
	place_first_strings();
	box_placement();
	placement_prep();
	modulesJustPlaced = partition_placement();
	/* Setup the bounding box */
	systerms = make_room_for_systerm_placement(&systermNets, &x1, &y1, &x2, &y2);

	fprintf(stderr,"\nAutomatic placement ");
    fprintf(stderr,"completed.\n");


    if (includeSysterms == TRUE)
    {
		/* routing */
		if (do_routing == TRUE) 
		{
			currentPartition = 0;
			for(ml = systerms; ml != NULL; ml = ml->next) pull_terms_from_nets(ml->object);
			routingRoot[VERT] = (tile **)calloc(partition_count + 2, sizeof(tile *));
			routingRoot[HORZ] = (tile **)calloc(partition_count + 2, sizeof(tile *));
			
			fprintf(stderr,"Routing for Systerm Placement...");
			create_routing_space(modulesJustPlaced, currentPartition, 
					 x_sizes[currentPartition], y_sizes[currentPartition], 
					 xfloor, yfloor);

			nl = first_global_route(modulesJustPlaced, currentPartition, 
						TRUE, congestion_rule);

			/* Create a local route to use to locate the system terminals */
			local_route(systermNets, currentPartition, FALSE);
			fprintf(stderr,"placing systerms...");

			/* Add the system terminals (systerms) to the diagram: */
			modulesJustPlaced = fine_systerm_placement(&x1, &y1, &x2, &y2); 
			fprintf(stderr,"completed. \n");
		}

		/* Add the system terminals back to their respective nets: */
		for(ml = systerms; ml != NULL; ml = ml->next) 
			add_terms_to_nets(ml->object); 
		fprintf(stderr,"System Terminals being added to diagram...\n");

		/* finish the routing */
		if (do_routing == TRUE)
		{
			fprintf(stderr,"Adding systerms to routing space, Global Routing commenced...");
			/* Deal with resetting the boundaries to include the system terminals: */
			reset_boundaries(routingRoot[HORZ][currentPartition], x1, y1, x2, y2);
			reset_boundaries(routingRoot[VERT][currentPartition], y1, x1, y2, x2);
			xfloor = x1;		x_sizes[currentPartition] = x2 - x1;
			yfloor = y1;		y_sizes[currentPartition] = y2 - y1;	    
			
			nl = incremental_global_route(modulesJustPlaced, modules, nl, currentPartition);

			fprintf(stderr,"completed...Local Routing commenced...");
			local_route(nets, currentPartition, TRUE);
			fprintf(stderr,"completed.\n");
		}
    }


    /*
     * finish up
     */


//	print_global_route(groutputFile);
	print_local_route(lroutputFile, nets);
//	ps_print_tile_border(lroutputFile,);


	fprintf(stderr,"\n %d Lines read, %d Modules, %d Nodes, %d Partitions\n",
		lcount, module_count, node_count, partition_count);
	fprintf(stderr,"\nMax partition size: %d Max Connections: %d Ratio: %-5.3f Rule #%d\n",
		max_partition_size, max_partition_conn, partition_ratio, partition_rule);
	fclose(ivffile);


	free_list(modules);
	free_list(nets); 
}

/* ------------------------------------------------------------------- */
int count_terminals(mlist *mods)
{
    /* Given a list of modules, count the number of terminals contained */
    int c = 0;
    mlist *ml;
    for (ml = mods; ml != NULL; ml = ml->next)
    {
	c += list_length(ml->object->terms);
    }
    return(c);
}

 
/*---------------------------------------------------------------
 * END OF FILE
 *---------------------------------------------------------------
 */







